@@ -24,6 +24,7 @@
 """
 Mu32 documentation is available on https://distalsense.io
 See documentation on usb device programming with libusb on https://pypi.org/project/libusb1/1.3.0/#documentation
+Examples are available on https://github.com/vpelletier/python-libusb1
 
 Please, note that the following packages should be installed before using this program:
 	> pip install libusb1
@@ -42,6 +43,9 @@ import numpy as np
 from math import ceil as ceil
 import queue
 
+from .log import mu32log
+from .exception import Mu32Exception
+
 MU_BEAMS_NUMBER				= 4
 MU_BEAM_MEMS_NUMBER			= 8
 MU_ANALOG_CHANNELS		    = 4
@@ -74,56 +78,6 @@ MU_DEFAULT_BUFFER_LENGTH		= 512
 MU_DEFAULT_DURATION				= 1
 
 
-class Mu32Formatter(logging.Formatter):
-	"""Logging Formatter to add colors and count warning / errors"""
-
-	green = "\x1b[32;21m"
-	blue = "\x1b[34;21m"
-	magenta = "\x1b[35;21m"
-	grey = "\x1b[38;21m"
-	yellow = "\x1b[33;21m"
-	red = "\x1b[31;21m"
-	bold_red = "\x1b[31;1m"
-	bold_black = "\x1b[30;1m"
-	reset = "\x1b[0m"
-	start_format = magenta + "%(asctime)s " + reset + bold_black + "[%(levelname)s]: " + reset
-
-	FORMATS = {
-		logging.DEBUG: start_format + green + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
-		logging.INFO: magenta + "%(asctime)s " + reset + "[%(levelname)s]: " + blue + "%(message)s" + reset,
-        logging.WARNING: start_format + yellow + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
-        logging.ERROR: start_format + red + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
-        logging.CRITICAL: start_format + bold_red + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset
-    }
-
-	def format(self, record):
-		log_fmt = self.FORMATS.get( record.levelno )
-		formatter = logging.Formatter( log_fmt )
-		return formatter.format( record )
-
-
-
-mu32ch = logging.StreamHandler()
-mu32ch.setLevel( logging.DEBUG )
-mu32ch.setFormatter( Mu32Formatter() )
-
-mu32log = logging.getLogger( __name__ )
-mu32log.setLevel( logging.WARNING )
-mu32log.addHandler( mu32ch )
-
-
-class Mu32Exception( Exception ):
-	value = None
-
-	def __init__( self, value=None ):
-		Exception.__init__( self )
-		if value is not None:
-			self.value = value
-
-	def __str__(self):
-		return '%s [%s]' % ( 'Mu32Exception: ', self.value )
-
-
 class MegaMicro:
 	pass
 
@@ -418,6 +372,17 @@ class Mu32( MegaMicro ):
 		self.ctrlWrite( handle, 0xB1, buf )
 
 
+	"""
+	Callback flushing function: only intended to flush Mu32 internal buffers
+	"""
+	def processFlush( self, transfer ):
+		if transfer.getActualLength() > 0:
+			mu32log.info( ' .flushed %d data bytes' % transfer.getActualLength() )
+
+
+	"""
+	Callback run function: check error transfer, call user callback function and then submit for new transfer
+	"""
 	def processRun( self, transfer ):
 		"""
 		Run callback function
@@ -449,7 +414,14 @@ class Mu32( MegaMicro ):
 		Otherwise push data in the object signal queue
 		"""
 		if self._callback_fn != None:
-			self._callback_fn( self, data )
+			try:
+				self._callback_fn( self, data )
+			except KeyboardInterrupt as e:
+				mu32log.critical( 'Mu32: keyboard interruption...' )
+				self._recording = False
+			except Exception:
+				mu32log.critical( 'Mu32: unexpected error: aborting... ' )
+				self._recording = False
 		else:
 			self._signal_q.put( data )
 		
@@ -577,6 +549,28 @@ class Mu32( MegaMicro ):
 						Send stop command to Mu32 FPGA
 						"""
 						self.ctrlStop( handle )
+
+						"""
+						Flush Mu32 remaining data 
+						"""
+						for transfer in transfer_list:
+							transfer.setBulk(
+								usb1.ENDPOINT_IN | MU_USB3_BUS_ADDRESS,
+								self._mems_number * MU_TRANSFER_DATAWORDS_SIZE * self._buffer_length,
+								callback=self.processFlush,
+								timeout=1
+							)
+							transfer.submit()
+
+						while any( x.isSubmitted() for x in transfer_list ):
+							try:
+								context.handleEvents()
+							except Exception as e:
+								raise Mu32Exception( 'Mu32 USB3 transfer flushing failed: [%s]' % e )
+
+						"""
+						Reset Mu32
+						"""
 						self.ctrlResetAcq32( handle )
 
 				except Exception as e:	
